// Cory McWilliams
// mcwilc@rpi.edu
// 10/15/2004

#include "Humanoid.h"

#include <stack>
#include "proj2_support.h"
#include <iostream>

#define HUMANOID_SPHERE_SLICES 20
#define HUMANOID_SPHERE_STACKS 20
#define HUMANOID_SPHERE_RADIUS 0.5f

#define ABS(a) (((a)>0)?(a):(-(a)))

Humanoid::Humanoid()
: mAnimation(0), mFrame(0), mTime(0.0f), mHolding(0) {
	for (int i = 0; i < HUMANOID_JOINT_COUNT; i++) {
		mCurrentRotations[i] = 0;
		mJoint[i] = 0;
	}
}

Humanoid::~Humanoid() {
	gluDeleteQuadric(mQuadric);
}

void Humanoid::draw() {
	glPushMatrix();
		glRotatef(mJoint[Joint_Hip], 1.0f, 0.0f, 0.0f);
		glPushMatrix();
		glScalef(HIP_WIDTH, HIP_HEIGHT, HIP_DEPTH);
		gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
		glPopMatrix();

		glPushMatrix();
			glRotatef(mJoint[Joint_RightUpperLeg], 1.0f, 0.0f, 0.0f);
			glTranslatef(HIP_WIDTH/2-UPPER_LEG_WIDTH/2, -HIP_HEIGHT/2-UPPER_LEG_HEIGHT/2, 0.0f);

			glPushMatrix();
			glScalef(UPPER_LEG_WIDTH, UPPER_LEG_HEIGHT, UPPER_LEG_DEPTH);
			gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
			glPopMatrix();

			glPushMatrix();
			glTranslatef(0.0f, -UPPER_LEG_HEIGHT/2, 0.0f);
			glRotatef(mJoint[Joint_RightLowerLeg], 1.0f, 0.0f, 0.0f);
			glTranslatef(0.0f, -LOWER_LEG_HEIGHT/2, 0.0f);
			glScalef(LOWER_LEG_WIDTH, LOWER_LEG_HEIGHT, LOWER_LEG_DEPTH);
			gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
			glPopMatrix();
		glPopMatrix();

		glPushMatrix();
			glRotatef(mJoint[Joint_LeftUpperLeg], 1.0f, 0.0f, 0.0f);
			glTranslatef(-HIP_WIDTH/2+UPPER_LEG_WIDTH/2, -HIP_HEIGHT/2-UPPER_LEG_HEIGHT/2, 0.0f);

			glPushMatrix();
			glScalef(UPPER_LEG_WIDTH, UPPER_LEG_HEIGHT, UPPER_LEG_DEPTH);
			gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
			glPopMatrix();

			glPushMatrix();
			glTranslatef(0.0f, -UPPER_LEG_HEIGHT/2, 0.0f);
			glRotatef(mJoint[Joint_LeftLowerLeg], 1.0f, 0.0f, 0.0f);
			glTranslatef(0.0f, -LOWER_LEG_HEIGHT/2, 0.0f);
			glScalef(LOWER_LEG_WIDTH, LOWER_LEG_HEIGHT, LOWER_LEG_DEPTH);
			gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
			glPopMatrix();
		glPopMatrix();

		glPushMatrix();
			glRotatef(mJoint[Joint_Torso], 1.0f, 0.0f, 0.0f);
			glTranslatef(0.0f, HIP_HEIGHT/2+TORSO_HEIGHT/2, 0.0f);

			glPushMatrix();
			glScalef(TORSO_WIDTH, TORSO_HEIGHT, TORSO_DEPTH);
			gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
			glPopMatrix();

			glPushMatrix();
				glTranslatef(0.0f, TORSO_HEIGHT/2, 0.0f);
				glRotatef(mJoint[Joint_Neck], 1.0f, 0.0f, 0.0f);
				glTranslatef(0.0f, NECK_HEIGHT/2, 0.0f);

				glPushMatrix();
				glScalef(NECK_WIDTH, NECK_HEIGHT, NECK_DEPTH);
				gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
				glPopMatrix();

				glPushMatrix();
				glRotatef(mJoint[Joint_Head], 1.0f, 0.0f, 0.0f);
				glTranslatef(0.0f, NECK_HEIGHT/2+HEAD_HEIGHT/2, 0.0f);
					glPushMatrix();
					glScalef(HEAD_WIDTH, HEAD_HEIGHT, HEAD_DEPTH);
					gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
					glPopMatrix();

					glPushMatrix();
					glTranslatef(HEAD_WIDTH/4, HEAD_HEIGHT/4, HEAD_DEPTH/2.0f);
					glScalef(0.4f, 0.4f, 0.4f);
					gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
					glPopMatrix();

					glPushMatrix();
					glTranslatef(-HEAD_WIDTH/4, HEAD_HEIGHT/4, HEAD_DEPTH/2.0f);
					glScalef(0.4f, 0.4f, 0.4f);
					gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
					glPopMatrix();
				glPopMatrix();
			glPopMatrix();

			glPushMatrix();
				glTranslatef(TORSO_WIDTH/2, TORSO_HEIGHT/2-SHOULDER_HEIGHT/2, 0.0f);
				glRotatef(mJoint[Joint_RightShoulder], 0.0f, 1.0f, 0.0f);

				glPushMatrix();
				glScalef(SHOULDER_WIDTH, SHOULDER_HEIGHT, SHOULDER_DEPTH);
				gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
				glPopMatrix();

				glTranslatef(SHOULDER_WIDTH/2, 0.0f, 0.0f);
				glRotatef(mJoint[Joint_RightUpperArm], 0.0f, 0.0f, 1.0f);
				glTranslatef(0.0f, UPPER_ARM_HEIGHT/2, 0.0f);

				glPushMatrix();
				glScalef(UPPER_ARM_WIDTH, UPPER_ARM_HEIGHT, UPPER_ARM_DEPTH);
				gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
				glPopMatrix();

				glTranslatef(0.0f, UPPER_ARM_HEIGHT/2, 0.0f);
				glRotatef(mJoint[Joint_RightLowerArm], 1.0f, 0.0f, 0.0f);
				glTranslatef(0.0f, LOWER_ARM_HEIGHT/2, 0.0f);

				glPushMatrix();
				glScalef(UPPER_ARM_WIDTH, UPPER_ARM_HEIGHT, UPPER_ARM_DEPTH);
				gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
				glPopMatrix();

			glPopMatrix();

			// Here's a little something extra...he can "hold" an object.
			if (mHolding) {
				glPushMatrix();
				glTranslatef(-TORSO_WIDTH/2, TORSO_HEIGHT/2-SHOULDER_HEIGHT/2, 0.0f);
				glRotatef(mJoint[Joint_LeftShoulder], 0.0f, 1.0f, 0.0f);

				glTranslatef(-SHOULDER_WIDTH/2, 0.0f, 0.0f);
				glRotatef(mJoint[Joint_LeftUpperArm], 0.0f, 0.0f, 1.0f);
				glTranslatef(0.0f, UPPER_ARM_HEIGHT/2, 0.0f);

				glTranslatef(0.0f, UPPER_ARM_HEIGHT/2, 0.0f);
				glRotatef(mJoint[Joint_LeftLowerArm], 1.0f, 0.0f, 0.0f);
				glTranslatef(0.0f, LOWER_ARM_HEIGHT, 0.0f);

				mHolding();
				glPopMatrix();
			}

			glPushMatrix();
				glTranslatef(-TORSO_WIDTH/2, TORSO_HEIGHT/2-SHOULDER_HEIGHT/2, 0.0f);
				glRotatef(mJoint[Joint_LeftShoulder], 0.0f, 1.0f, 0.0f);

				glPushMatrix();
				glScalef(SHOULDER_WIDTH, SHOULDER_HEIGHT, SHOULDER_DEPTH);
				gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
				glPopMatrix();


				glTranslatef(-SHOULDER_WIDTH/2, 0.0f, 0.0f);
				glRotatef(mJoint[Joint_LeftUpperArm], 0.0f, 0.0f, 1.0f);
				glTranslatef(0.0f, UPPER_ARM_HEIGHT/2, 0.0f);

				glPushMatrix();
				glScalef(UPPER_ARM_WIDTH, UPPER_ARM_HEIGHT, UPPER_ARM_DEPTH);
				gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
				glPopMatrix();

				glTranslatef(0.0f, UPPER_ARM_HEIGHT/2, 0.0f);
				glRotatef(mJoint[Joint_LeftLowerArm], 1.0f, 0.0f, 0.0f);
				glTranslatef(0.0f, LOWER_ARM_HEIGHT/2, 0.0f);

				glPushMatrix();
				glScalef(UPPER_ARM_WIDTH, UPPER_ARM_HEIGHT, UPPER_ARM_DEPTH);
				gluSphere(mQuadric, HUMANOID_SPHERE_RADIUS, HUMANOID_SPHERE_SLICES, HUMANOID_SPHERE_STACKS);
				glPopMatrix();

			glPopMatrix();
		glPopMatrix();
	glPopMatrix();
}

void Humanoid::loopAnimation(const char *filename) {
	delete mAnimation;
	mAnimation = 0;

	mAnimation = new animation_handler(filename);
	mFrame = 0;
	mTime = 0.0f;
}

void Humanoid::step() {
	if (mAnimation && mAnimation->size()) {
		const int frames = mAnimation->size();
		mTime += 0.1f;
		if (mTime > 1.0f) {
			mFrame++;
			mTime -= 1.0f;
			if (static_cast<size_t>(mFrame) >= mAnimation->size()) mFrame = 0;
		}

		int v[4] = {(mFrame + frames - 1) % frames, mFrame, (mFrame + frames + 1) % frames, (mFrame + frames + 2) % frames};

		for (int i = 0; i < HUMANOID_JOINT_COUNT-1; i++) {
			mJoint[i] = catmullRomAngle((*mAnimation)[v[0]][i], (*mAnimation)[v[1]][i], (*mAnimation)[v[2]][i], (*mAnimation)[v[3]][i], mTime);
		}
	}
}

/*
	This function just adds or subtracts 360 degrees to the second argument
	so that the arguments represent angles in the nearest form to one another.
	This makes interpolation work right and not make excessive turns or go the
	wrong way.
*/
void nearAngle(const float a, float &b) {
	if (b - a > 180.0f) b -= 360.0f;
	else if (a - b > 180.0f) b += 360.0f;
}

float Humanoid::catmullRomAngle(float v1, float v2, float v3, float v4, float t) {
	nearAngle(v1, v2);
	nearAngle(v2, v3);
	nearAngle(v3, v4);
	return 0.5f * ((2 * v2) + (-1.0f * v1 + v3) * t + (2.0f * v1 - 5.0f * v2 + 4.0f * v3 - v4) * t * t + (-1.0f * v1 + 3.0f * v2 - 3.0f * v3 + v4) * t * t * t);
}

void Humanoid::dump() {
	for (int i = 0; i < 13; i++) {
		if (i == 13 - 1) {
			std::cout << mJoint[i] << '\n';
		} else {
			std::cout << mJoint[i] << ' ';
		}
	}
}

void Humanoid::reset() {
	delete mAnimation;
	mAnimation = 0;

	for (int i = 0; i < HUMANOID_JOINT_COUNT; i++) {
		mJoint[i] = 0.0f;
	}

	mFrame = 0;
	mTime = 0.0f;

	if (!mQuadric) {
		mQuadric = gluNewQuadric();
	}

}

bool Humanoid::lastFrame() {
	return !mAnimation || static_cast<size_t>(mFrame) == mAnimation->size() - 1;
}

void Humanoid::hold(void (*holding)()) {
	mHolding = holding;
}
